package asdraw;

import asdraw.World;

import java.awt.event.KeyAdapter;
import java.awt.event.KeyEvent;


/**
 * <p>The implementation of callbacks for the key events.</p>
 * <p>Report all regular key presses and the four arrow keys</p>
 * <p>Ignore other key pressed events, key released events, 
 * special keys, etc.</p>
 * 
 * @author Viera K. Proulx
 * @since August 2, 2007
 */
class MyKeyAdapter extends KeyAdapter {

  /** the current <code>{@link World World}</code> 
   * that handles the key events */
  protected World currentWorld;

  /** the <code>KeyAdapter</code> that handles the key events */
  protected MyKeyAdapter(World currentWorld){
    this.currentWorld = currentWorld;
  }

  /**
   * The callback for the key pressed event
   */
  protected void keyEventCallback(String keyString){
    this.currentWorld.processKeyEvent(keyString); 
  }

  /**
   * The callback for the key released event
   */
  protected void keyReleasedEventCallback(String keyString){
    this.currentWorld.processKeyReleasedEvent(keyString); 
  }

  // --------------------------------------------------------------------//
  // the key event handlers                                              //
  // --------------------------------------------------------------------//-

  /** 
   * <p>Handle the key typed event from the canvas.</p>
   * <p>This is where we get the letter keys.</p>
   *
   * @param e the <code>KeyEvent</code> that caused the callback
   */
  public void keyTyped(KeyEvent e) {
    displayInfo(e, "KEY TYPED: ");
  }

  /** 
   * <p>Handle the key pressed event from the canvas.</p>
   * <p>This is where we get the arrow keys.</p>
   *
   * @param e the <code>KeyEvent</code> that caused the callback
   */  
  public void keyPressed(KeyEvent e) {
    displayInfo(e, "KEY PRESSED: ");
  }
  
  /** 
   * <p>Handle the key typed event from the canvas.</p>
   * <p>This is where we get the letter keys.</p>
   *
   * @param e the <code>KeyEvent</code> that caused the callback
   */
  public void keyReleased(KeyEvent e) {
    displayInfo(e, "KEY RELEASED: ");
  }


  /**
   * <p>The key event major processor. The code is adopted from the code
   * provided by the Java documentation for key event processing.</p>
   * 
   * <p><em>We have to jump through some hoops to avoid
   * trying to print non-printing characters 
   * such as Shift.  (Not only do they not print, 
   * but if you put them in a String, the characters
   * afterward won't show up in the text area.)</em></P>
   * 
   * <p>There may be unneeded code here. An earlier comment stated:
   * <em>needs to add conversion table from code to 
   * </em><code>String</code>. The code for converting a single character
   * to <code>String</code> is included here.</p>
   */
  protected void displayInfo(KeyEvent e, String s){
    String keyString ="";
    String modString, tmpString,
    actionString, locationString;

    //You should only rely on the key char if the event
    //is a key pressed or released event.
    int id = e.getID();
    //if (id == KeyEvent.KEY_TYPED) {
      // regular letter has been pressed and released
      //char c = e.getKeyChar();
      //keyString = "" + c;

      // process four arrow keys when pressed
    if (//id == KeyEvent.KEY_TYPED ||
        id == KeyEvent.KEY_PRESSED ||
        id == KeyEvent.KEY_RELEASED){
      
      // check if pressed is one of the arrows
      int keyCode = e.getKeyCode();
      if (keyCode == 37)
        keyString = "left"; 
      else if (keyCode == 38)
        keyString = "up"; 
      else if (keyCode == 39)
        keyString = "right"; 
      else if (keyCode == 40)
        keyString = "down"; 

      else{
        // regular letter has been pressed or released
        char c = e.getKeyChar();
        keyString = "" + c;
      } 
/*
      else{
        keyString = KeyEvent.getKeyText(e.getKeyCode());
      }
*/
    }

    /*
    if (id == KeyEvent.KEY_RELEASED)
      System.out.println("Key " + keyString + 
          " keyCode " //+ e.getKeyCode() + " event: " 
          + "released");
    */
    /*
    if (id == KeyEvent.KEY_TYPED)
      System.out.println("Key " + keyString + 
          " keyCode " + e.getKeyCode() + " event: " + "typed");
    */
    /*
    if (id == KeyEvent.KEY_PRESSED)
      System.out.println("Key " + keyString + 
          " keyCode " //+ e.getKeyCode() + " event: " 
          + "pressed");
    */
    // ignore other keys pressed

    //////////////////////////////////////////////////////////////////////////
    // here are the actual callbacks for key pressed and key released                                         //
    if (keyString.length() != 0)
      
       if (id == KeyEvent.KEY_PRESSED) 
       //if (id == KeyEvent.KEY_TYPED) 
           keyEventCallback(keyString); 
    
       else if (id == KeyEvent.KEY_RELEASED) 
            keyReleasedEventCallback(keyString);
     
    //
    //////////////////////////////////////////////////////////////////////////
  } 
  // ---------------------------------------------------------------------
  // end of the key event major processor
}
