package idraw;

import geometry.*;

import java.awt.Insets;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.KeyAdapter;
import java.awt.event.KeyEvent;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;

import javax.swing.Timer;

import idraw.MyKeyAdapter;
import idraw.MyTimer;

/**
 * Copyright 2007, 2008, 2011 Viera K. Proulx
 * This program is distributed under the terms of the 
 * GNU Lesser General Public License (LGPL)
 */


/**
 * World for programming interactive games - with graphics, key events,
 * mouse events and a timer. Designed to implement the same interface as
 * DrScheme/ProfessorJ world teachpack library.
 * 
 * @author Viera K. Proulx
 * @since November 15, 2007, 14 July 2011
 */
abstract public class World{

  /** the canvas that displays the current world */
  public Canvas theCanvas;

  /** true if 'bigBang' started the world and it did not end, did not stop */
  private transient boolean worldExists = false;

  /** the timer for this world */
  protected transient MyTimer mytime;
  
  /** timer events not processed when the mouse event is processed */
  protected transient boolean stopTimer = false;

  /** the key adapter for this world*/
  private transient MyKeyAdapter ka;
  
  /** the mouse adapter for this world */
  private transient MyMouseAdapter ma;
 
  /**
   * The default constructor. To start the world one must invoke the
   * <code>bigBang</code> method.
   */
  public World(){ }

  /////////////////////////////////////////////////////////////////////////
  // Methods for interacting with the World                              //
  /////////////////////////////////////////////////////////////////////////

  /**
   * Start the world by creating a canvas of the given size, creating
   * and adding the key and mouse adapters, and starting the timer at the
   * given speed.
   * 
   * @param w the width of the <code>{@link Canvas Canvas}</code>
   * @param h the height of the <code>{@link Canvas Canvas}</code>
   * @param speed the speed at which the clock runs
   */
  public void bigBang(int w, int h, double speed){
    // throw runtime exceptions if w, h <= 0
    this.theCanvas = new Canvas(w, h);
    this.worldExists = true;
    
    // pause a bit so that two canvases do not compete when being opened
    // almost at the same time
    long start = System.currentTimeMillis();
    long tmp = System.currentTimeMillis();
    // System.out.println("Going to sleep.");
	
    while(tmp - start < 1000){
    	tmp = System.currentTimeMillis();
    }

    // add the key listener to the frame for our canvas
    this.ka = new MyKeyAdapter(this);
    this.theCanvas.f.addKeyListener(this.ka);
    
    // add the mouse listener to the frame for our canvas
    this.ma = new MyMouseAdapter(this);
    this.theCanvas.f.addMouseListener(this.ma);

    // make sure the canvas responds to events
    this.theCanvas.f.setFocusable(true);

    // add the timer -- start it if speed is greater than 0
    this.mytime = new MyTimer(this, speed);
    if (speed > 0.0)
      this.mytime.timer.start();

    // finally, show the canvas and draw the initial world
    this.theCanvas.show();

    // pause again after the Canvas is shown to make sure 
    // all listeners and the timer are installed for theCanvas
    start = System.currentTimeMillis();
    tmp = System.currentTimeMillis();
    // System.out.println("Going to sleep again.");
	
    while(tmp - start < 1000){
    	tmp = System.currentTimeMillis();
    }
    
    this.drawWorld();
  }


  /**
   * Stop the timer, disable event processing - leave the canvas open, 
   * write the end of time message onto the canvas
   */
  public void endOfTime(String s){
    if (this.worldExists){
      // remove listeners and set worldExists to false
      this.mytime.timer.stop();
      this.worldExists = false;
      this.mytime.stopTimer();
      this.theCanvas.f.removeKeyListener(this.ka);
      System.out.println("The world stopped: " + s);

      // draw the final scene of the world with the end of time message
      this.theCanvas.clear();
      this.draw();
      this.theCanvas.drawString(new Posn(10, 20),s);
    }
  }


  /**
   * End the world interactions - leave the canvas open, 
   * write the end of world message onto the canvas
   */
  public void endOfWorld(String s){
    if (this.worldExists){
      // remove listeners and set worldExists to false
      this.mytime.timer.stop();
      this.worldExists = false;
      this.mytime.stopTimer();
      this.theCanvas.f.removeKeyListener(this.ka);
      System.out.println("The end of the world: " + s);

      // draw the final scene of the world with the end of world message
      this.theCanvas.clear();
      this.draw();
      this.theCanvas.drawString(new Posn(10, 20),s);
    }
  }

  /**
   * The method invoked by the timer on each tick.
   * Delegates to the user to define a new state of the world,
   * then resets the canvas and event handlers for the new world
   * to those currently used.
   * 
   * @return <code>{@link World World}</code> after the tick event
   */
  protected void processTick(){
    try{
      if (this.worldExists && !this.stopTimer){
        this.onTick();
        this.drawWorld();
      }
    } 
    catch(RuntimeException re){
      re.printStackTrace();
      endOfTime(re.getMessage());
      //throw re;
      Runtime.getRuntime().halt(1);
    }
  }  

  /**
   * <P>User defined method to be invoked by the timer on each tick.
   * Update the <code>{@link World World}</code>.</P>
   * <P>Override this method in the game world class</P>
   */
  abstract public void onTick();
    

  /**
   * The method invoked by the key adapter on selected key events.
   * Delegates to the user to define a new state of the world,
   * then resets the canvas and event handlers for the new world
   * to those currently used.
   * 
   * @return <code>{@link World World}</code> after the key event
   */
  protected void processKeyEvent(String ke){
    try{
      if (this.worldExists){
        this.onKeyEvent(ke);
        this.drawWorld();
      }
    }
    catch(RuntimeException re){
      re.printStackTrace();
      endOfTime(re.getMessage());
      //throw re;
      Runtime.getRuntime().halt(1);
    }
  }

  /**
   * <P>User defined method to be invoked by the key adapter 
   * on selected key events.
   * Update the <code>{@link World World}</code>.</P>
   * <P>Override this method in the game world class</P>
   */
  abstract public void onKeyEvent(String s);
  
  /**
   * The method invoked by the mouse adapter on mouse clicked event.
   * Delegates to the user to define a new state of the world.
   * 
   * @param mouse the location of the mouse when clicked
   */
  protected void processMouseClicked(Posn mouse){
    try{
      if (this.worldExists){

        // process the changes to the world on this mouse clicked event
        this.onMouseClicked(mouse);
        
        // draw the world as it looks after this mouse event
        this.drawWorld();
      }
    }
    catch(RuntimeException re){
      re.printStackTrace();
      endOfTime(re.getMessage());
      //throw re;
      Runtime.getRuntime().halt(1);
    }
  }

  /**
   * <P>User defined method to be invoked by the mouse adapter 
   * when a mouse is clicked.
   * Update the <code>{@link World World}</code>.</P>
   * <P>Override this method in the game world class</P>
   * 
   * @param mouse the location of the mouse when clicked
   */
  public void onMouseClicked(Posn mouse){}
  
  /**
   * The method invoked by the mouse adapter on mouse entered event.
   * Delegates to the user to define a new state of the world.
   * 
   * @param mouse the location of the mouse when entered
   */
  protected void processMouseEntered(Posn mouse){
    try{
      if (this.worldExists){

        // process the changes to the world on this mouse entered event
        this.onMouseEntered(mouse);
        
        // draw the world as it looks after this mouse event
        this.drawWorld();
      }
    }
    catch(RuntimeException re){
      re.printStackTrace();
      endOfTime(re.getMessage());
      //throw re;
      Runtime.getRuntime().halt(1);
    }
  }

  /**
   * <P>User defined method to be invoked by the mouse adapter 
   * when a mouse is entered.
   * Update the <code>{@link World World}</code>.</P>
   * <P>Override this method in the game world class</P>
   * 
   * @param mouse the location of the mouse when entered
   */
  public void onMouseEntered(Posn mouse){}  
  
  
  /**
   * The method invoked by the mouse adapter on mouse exited event.
   * Delegates to the user to define a new state of the world.
   * 
   * @param mouse the location of the mouse when exited
   */
  protected void processMouseExited(Posn mouse){
    try{
      if (this.worldExists){

        // process the changes to the world on this mouse exited event
        this.onMouseExited(mouse);
        
        // draw the world as it looks after this mouse event
        this.drawWorld();
      }
    }
    catch(RuntimeException re){
      re.printStackTrace();
      endOfTime(re.getMessage());
      //throw re;
      Runtime.getRuntime().halt(1);
    }
  }

  /**
   * <P>User defined method to be invoked by the mouse adapter 
   * when a mouse is exited.
   * Update the <code>{@link World World}</code>.</P>
   * <P>Override this method in the game world class</P>
   * 
   * @param mouse the location of the mouse when exited
   */
  public void onMouseExited(Posn mouse){}  
  
  
  /**
   * The method invoked by the mouse adapter on mouse pressed event.
   * Delegates to the user to define a new state of the world.
   * 
   * @param mouse the location of the mouse when pressed
   */
  protected void processMousePressed(Posn mouse){
    try{
      if (this.worldExists){

        // process the changes to the world on this mouse pressed event
        this.onMousePressed(mouse);
        
        // draw the world as it looks after this mouse event
        this.drawWorld();
      }
    }
    catch(RuntimeException re){
      re.printStackTrace();
      endOfTime(re.getMessage());
      //throw re;
      Runtime.getRuntime().halt(1);
    }
  }

  /**
   * <P>User defined method to be invoked by the mouse adapter 
   * when a mouse is pressed.
   * Update the <code>{@link World World}</code>.</P>
   * <P>Override this method in the game world class</P>
   * 
   * @param mouse the location of the mouse when pressed
   */
  public void onMousePressed(Posn mouse){}
  
  /**
   * The method invoked by the mouse adapter on mouse released event.
   * Delegates to the user to define a new state of the world.
   * 
   * @param mouse the location of the mouse when released
   */
  protected void processMouseReleased(Posn mouse){
    try{
      if (this.worldExists){

        // process the changes to the world on this mouse released event
        this.onMouseReleased(mouse);
        
        // draw the world as it looks after this mouse event
        this.drawWorld();
      }
    }
    catch(RuntimeException re){
      re.printStackTrace();
      endOfTime(re.getMessage());
      //throw re;
      Runtime.getRuntime().halt(1);
    }
  }
  
  /**
   * <P>User defined method to be invoked by the mouse adapter 
   * when a mouse is released.
   * Update the <code>{@link World World}</code>.</P>
   * <P>Override this method in the game world class</P>
   * 
   * @param mouse the location of the mouse when released
   */
  public void onMouseReleased(Posn mouse){}

  /**
   * Invoke the user defined <code>draw</code> method, if this 
   * <code>{@link World World}</code> has been initialized 
   * via <code>bigBang</code> and did not stop or end 
   * via <code>endOfTime</code> 
   * or <code>endOfWorld</code>
   */
  protected void drawWorld(){
    if (this.worldExists){
      this.theCanvas.clear();
      this.draw();
    }
  }

  /**
   * <P>User defined method to draw the <code>{@link World World}</code>.</P>
   * <P>Override this method in the game world class</P>
   */
  abstract public void draw();
}

/**
 * The action listener for the timer events.
 * 
 * @author Viera K. Proulx
 * @since August 2, 2007
 */
class MyTimer{

  /** the current <code>{@link World World}</code> 
   * that handles the timer events */
  protected World currentWorld;

  /** the <code>Timer</code> that generates the time events */
  protected Timer timer;
  
  public boolean running = true;

  /** the timer speed */
  protected int speed;

  /**
   * Create the initial timer for the given 
   * <code>{@link World World}</code> at the given <code>speed</code>.
   * 
   * @param currentWorld the given <code>{@link World World}</code>
   * @param speed the given <code>speed</code>
   */
  protected MyTimer(World currentWorld, double speed){
    this.currentWorld = currentWorld;
    this.timer = new Timer((new Double(speed * 1000)).intValue(), 
        this.timerTasks);
    this.speed = (new Double(speed * 1000)).intValue();
  }

  /**
   * The callback for the timer events
   */
  protected ActionListener timerTasks = new ActionListener() {
    public void actionPerformed(ActionEvent evt) {
      if (running)
        currentWorld.processTick();
    }
  };

  /**
   * A helper method to convert the <code>speed</code> given as 
   * a delay time into milliseconds
   */
  protected void setSpeed(){ 
    this.timer.setDelay(this.speed);
  }
  
  protected void stopTimer(){
    this.running = false;
  }
}

/**
 * <p>The implementation of callbacks for the key events.</p>
 * <p>Report all regular key presses and the four arrow keys</p>
 * <p>Ignore other key pressed events, key released events, 
 * special keys, etc.</p>
 * 
 * @author Viera K. Proulx
 * @since August 2, 2007
 */
class MyKeyAdapter extends KeyAdapter {

  /** the current <code>{@link World World}</code> 
   * that handles the key events */
  protected World currentWorld;

  /** the <code>KeyAdapter</code> that handles the key events */
  protected MyKeyAdapter(World currentWorld){
    this.currentWorld = currentWorld;
  }

  /**
   * The callback for the key events
   */
  protected void keyEventCallback(String keyString){
    this.currentWorld.processKeyEvent(keyString); 
  }

  // --------------------------------------------------------------------//
  // the key event handlers                                              //
  // --------------------------------------------------------------------//-

  /** 
   * <p>Handle the key typed event from the canvas.</p>
   * <p>This is where we get the letter keys.</p>
   *
   * @param e the <code>KeyEvent</code> that caused the callback
   */
  public void keyTyped(KeyEvent e) {
    displayInfo(e, "KEY TYPED: ");
  }

  /** 
   * <p>Handle the key pressed event from the canvas.</p>
   * <p>This is where we get the arrow keys.</p>
   *
   * @param e the <code>KeyEvent</code> that caused the callback
   */  
  public void keyPressed(KeyEvent e) {
    displayInfo(e, "KEY PRESSED: ");
  }

  /**
   * <p>The key event major processor. The code is adopted from the code
   * provided by the Java documentation for key event processing.</p>
   * 
   * <p><em>We have to jump through some hoops to avoid
   * trying to print non-printing characters 
   * such as Shift.  (Not only do they not print, 
   * but if you put them in a String, the characters
   * afterward won't show up in the text area.)</em></P>
   * 
   * <p>There may be unneeded code here. An earlier comment stated:
   * <em>needs to add conversion table from code to 
   * </em><code>String</code>. The code for converting a single character
   * to <code>String</code> is included here.</p>
   */
  protected void displayInfo(KeyEvent e, String s){
    String keyString ="";
    String modString, tmpString,
    actionString, locationString;

    //You should only rely on the key char if the event
    //is a key typed event.
    int id = e.getID();
    if (id == KeyEvent.KEY_TYPED) {
    	// regular letter has been pressed and released
    	char c = e.getKeyChar();
    	keyString = "" + c;

    	// process four arrow keys when pressed
    } else if (id == KeyEvent.KEY_PRESSED){
    	// check if pressed is one of the arrows
    	int keyCode = e.getKeyCode();
    	if (keyCode == 37)
    		keyString = "left"; 
    	else if (keyCode == 38)
    		keyString = "up"; 
    	else if (keyCode == 39)
    		keyString = "right"; 
    	else if (keyCode == 40)
    		keyString = "down"; 

    	else
    		; // ignore other keys pressed

      // record the KEY_RELEASED event
    } else {
      keyString = "released";
    }

    //////////////////////////////////////////////////////////////////////////
    // here is the actual callback                                          //
    if (!(keyString.length() == 0 ||    // ignore key pressed except arrows //
        keyString.equals("released")))  // ignore key released              //
      keyEventCallback(keyString);                                          //
    //////////////////////////////////////////////////////////////////////////
  } 
  // ---------------------------------------------------------------------
  // end of the key event major processor
}


/**
 * <p>The implementation of callbacks for the mouse events: 
 * mouse clicked, entered, exited, pressed, and released.</p>
 * <p>We do not handle mouse motion events: mouse dragged and 
 * mouse moved.</p>
 * 
 * @author Viera K. Proulx
 * @since August 2, 2007
 */
class MyMouseAdapter extends MouseAdapter {

  /** the current <code>{@link World World}</code> 
   * that handles the mouse events */
  protected World currentWorld;

  /** the mouse position recorded by the mouse event handler */
  protected Posn mousePosn;
  
  /**
   * Create the mouse listener for the given <code>{@link World World}</code>.
   * 
   * @param currentWorld the given <code>{@link World World}</code>
   */
  protected MyMouseAdapter(World currentWorld){
    this.currentWorld = currentWorld;
  }
  
  // ---------------------------------------------------------------------
  // the mouse event handlers
  
  /**
   * Adjust the reported mouse position to account for the top bar
   * 
   * @param mousePosn the recorded mouse position
   * @return the actual mouse position
   */ 
  Posn adjustMousePosn(Posn mousePosn){
    // .... use this to find the height of the top bar
    Insets ins = this.currentWorld.theCanvas.f.getInsets();
    mousePosn.y -= ins.top;
    return mousePosn;
  }
   
  /**
   * Invoked when the mouse has been clicked on a component.
   * 
   * @param e the mouse event that invoked this callback
   */
  public void mouseClicked(MouseEvent e) {
    this.currentWorld.stopTimer = true;
    this.mousePosn = new Posn(e.getX(), e.getY());
    this.currentWorld.processMouseClicked(adjustMousePosn(this.mousePosn));
    this.currentWorld.stopTimer = false;
  }
  
  /**
   * Invoked when the mouse enters a component.
   * 
   * @param e the mouse event that invoked this callback
   */
  public void mouseEntered(MouseEvent e) {
    this.currentWorld.stopTimer = true;
    this.mousePosn = new Posn(e.getX(), e.getY());
    this.currentWorld.processMouseEntered(adjustMousePosn(this.mousePosn));
    this.currentWorld.stopTimer = false;
  }
  
  /**
   * Invoked when the mouse exits a component.
   * 
   * @param e the mouse event that invoked this callback
   */
  public void mouseExited(MouseEvent e) {
    this.currentWorld.stopTimer = true;
    this.mousePosn = new Posn(e.getX(), e.getY());
    this.currentWorld.processMouseExited(adjustMousePosn(this.mousePosn));
    this.currentWorld.stopTimer = false;
  }
  
  /**
   * Invoked when the mouse button has been pressed on a component.
   * 
   * @param e the mouse event that invoked this callback
   */
  public void mousePressed(MouseEvent e) {
    this.currentWorld.stopTimer = true;
    this.mousePosn = new Posn(e.getX(), e.getY());
    this.currentWorld.processMousePressed(adjustMousePosn(this.mousePosn));
    this.currentWorld.stopTimer = false;
  }
  
  /**
   * Invoked when a mouse button has been released on a component.
   * 
   * @param e the mouse event that invoked this callback
   */
  public void mouseReleased(MouseEvent e) {
    this.currentWorld.stopTimer = true;
    this.mousePosn = new Posn(e.getX(), e.getY());
    this.currentWorld.processMouseReleased(adjustMousePosn(this.mousePosn));
    this.currentWorld.stopTimer = false;
   }
}


